/*
 * Copyright (c) 2006,2007,2008 Jon Siddle <jon@trapdoor.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * 2008-07-16
 * - Fixed overzealous truncation caused by checking track length prior to
 *   sample size compensation.
 * 2008-04-22
 * - Fixed 64bit issues. Should be 64bit clean now
 * - Fixed the size of some variables / format specifiers
 * - Added Mac OS X ioctl for raw devices
 * - Moved variable definition inside ifdef to prevent compiler warning
 * - Added return value checks for ioctl, lseek
 * - Retabbing and Minor reformatting
 * - Added fallback block device size detection for non-linux systems
 * - Fixed really stupid bug rendering 16bit formats unplayable
 * 2008-01-16
 * - Improved and refactored path parsing [MK]
 * - Fixed a couple of minor bugs
 * - Renamed poorly named escape function
 * - Moved macros and types to header file
 * - Removed excessively lazy track initialisation
 * - Improved program presence detection
 * - Adapted code from [MK] to pre-load segment table
 *   upon open()
 * 2008-01-02
 * - Integrated commandline fix [MK]
 * - Added verbosity flag [MK]
 */

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#ifdef __linux__
#   include <linux/fs.h>
#endif
#ifdef __APPLE__
#   include <sys/disk.h>
#endif
#include <ctype.h>
#include <assert.h>

#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>

#include <inttypes.h>

#include "fdmsfs.h"

//Commandline long options
int lopt_flag = 0;
#define LOPT_BITS 1
#define LOPT_RATE 2
#define LOPT_HELP 'h'
#define LOPT_VERB 'v'

struct option long_options[] = {
    {"bits",1,&lopt_flag,LOPT_BITS},
    {"rate",1,&lopt_flag,LOPT_RATE},
    {"help",0,0,LOPT_HELP},
    {"verbose",0,0,LOPT_VERB},
    {0,0,0,0}
};

int verbose = 1;

static char wavheader[] = {
    'R', 'I', 'F', 'F',
    0, 0, 0, 0,			//file size
    'W', 'A', 'V', 'E',
    'f', 'm', 't', ' ',
    16, 0, 0, 0,		//size of tag
    1, 0,			//Format
    1, 0,			//Channels
    0x80, 0xbb, 0, 0,		//Samplerate 48000
    0x00, 0xee, 0x2, 0,		//avg byte/sec 48000*4
    4, 0,			//Block align
    32, 0,			//bits/sample
    'd', 'a', 't', 'a',
    0, 0, 0, 0			//size of tag
};

/*
 * Clobber troublesome (for filesystem) chars.
 */
static inline void fdmsfs_clobber(char *str) {
    while((*str)!='\0') {
	if(!( isalpha(*str) ||
	      isdigit(*str) ||
	      isspace(*str) ||
	      (*str)=='-'))
	    *str = '_';
	str++;
    }
}

/*
 * Parse program part of path and extract program number.
 */
static int parse_prg(const char *path)
{
    int p=-1;
    int n=0;

    sscanf(path,"/p%d-%n", &p, &n);

    if(p>=0 && p<128 && fdmsfs.prgs[p].present) {
	/*Copy name, and strip invalid chars*/
	char *name = strdup(fdmsfs.prgs[p].name);
	char *ptr = name;
	fdmsfs_clobber(ptr);

	if(!strcmp(name,path+n)) {
	    free(name);
	    return p;
	}
	free(name);
    }

    return -1;
}

void init_tracks(int prg)
{
    int i;

    for(i=0;i<24;i++) {
	int s;

	/* Seek to track's segment table */
	if( (lseek(fdmsfs.fd, ( fdmsfs.prgs[prg].offset+0x20+(8*i) )*0x200LL, SEEK_SET)) <0) {
	    perror("Seek error on device/image");
	    return;
	}

	/* Read segment table */
	for(s=0;s<0x200;s++){
	    uint32_t s_offset;
	    uint32_t s_length;

	    if(read(fdmsfs.fd,&s_offset,4)<0){
		perror("Error reading file");
		exit(1);
	    }
	    if(read(fdmsfs.fd,&s_length,4)<0){
		perror("Error reading file");
		exit(1);
	    }

	    SWAP_32(&s_offset);
	    SWAP_32(&s_length);

	    /* Stop at first 0 segment. */
	    if(s_offset==0 && s_length==0)
		break;

	    /* Stop if offset > file size */
	    if( s_offset*0x200LL >= fdmsfs.dev_size ) {
		dprintf( "Track offset beyond end of file\n" );
		break;
	    }

	    /* Truncate length if > file size */
	    if( (s_offset+s_length)*0x200LL > fdmsfs.dev_size ) {
		dprintf( "Track end beyond end of file, truncating\n" );
		s_length = fdmsfs.dev_size - s_offset*0x200LL;
		s_length /= 0x200LL;
	    }

	    fdmsfs.prgs[prg].tracks[i].refs = 0;
	    fdmsfs.prgs[prg].tracks[i].segments = NULL;
	    fdmsfs.prgs[prg].tracks[i].present = 1;
	    fdmsfs.prgs[prg].tracks[i].length += s_length;
	    dprintf( "track %d, segment @%" PRIx32 "\n",i, s_offset);
	}
    }
}

void init_segment_table(int prg, int trk)
{
    int s;

    /* Increase reference count */
    fdmsfs.prgs[prg].tracks[trk].refs++;

    /* Someone else got here first, no need to create table. */
    if(fdmsfs.prgs[prg].tracks[trk].refs>1)
	return;

    /* Seek to track's segment table */
    if( (lseek(fdmsfs.fd, ( fdmsfs.prgs[prg].offset+0x20+(8*trk) )*0x200LL, SEEK_SET)) <0) {
	perror("Seek error on device/image");
	exit(1);
    }

    /* Allocate segment table */
    if( (
	 fdmsfs.prgs[prg].tracks[trk].segments =
	 (struct fdmsfs_segment *)calloc( 0x200, sizeof(struct fdmsfs_segment) )
	)==NULL) {
	perror("Out of memory");
	exit(1);
    }

    /* Read segment table */
    for(s=0;s<0x200;s++){
	uint32_t s_offset;
	uint32_t s_length;

	if(read(fdmsfs.fd,&s_offset,4)<0){
	    perror("Error reading file");
	    exit(1);
	}
	if(read(fdmsfs.fd,&s_length,4)<0){
	    perror("Error reading file");
	    exit(1);
	}

	SWAP_32(&s_offset);
	SWAP_32(&s_length);

	/* Stop at first 0 segment. */
	if(s_offset==0 && s_length==0)
	    break;

	/* Stop if offset > file size */
	if( s_offset*0x200LL >= fdmsfs.dev_size ) {
	    dprintf( "Track offset beyond end of file\n" );
	    break;
	}

	/* Truncate length if > file size */
	if( (s_offset+s_length)*0x200LL > fdmsfs.dev_size ) {
	    dprintf( "Track end beyond end of file, truncating\n" );
	    s_length = fdmsfs.dev_size - s_offset*0x200LL;
	    s_length /= 0x200LL;
	}

	fdmsfs.prgs[prg].tracks[trk].segments[s].offset = s_offset;
	fdmsfs.prgs[prg].tracks[trk].segments[s].length = s_length;
    }
}

static void free_segment_table(int prg, int trk)
{
    /* Descrease reference count */
    fdmsfs.prgs[prg].tracks[trk].refs--;

    /* We were the last, free */
    if(!fdmsfs.prgs[prg].tracks[trk].refs) {
	free(fdmsfs.prgs[prg].tracks[trk].segments);
	fdmsfs.prgs[prg].tracks[trk].segments = NULL;
    }
}

/*
 * Parse path and extract program and track numbers.
 */
static enum fdmsfs_pathtype parse_path(const char *path, int *prg, int *trk) {
    char *prg_path = NULL;
    const char *ptr = path+1;

    /* --- Check path, extract track number, etc --- */

    if(!strcmp(path, "/"))
	return pt_root;

    /* Find / after prog, or \0
     * This emulates strchrnul()
     */
    ptr=strchr(ptr,'/');
    if(ptr==NULL) {
	ptr = path+strlen(path);
    }

    /* Copy path, truncate to the prog part, and find prog index */
    prg_path = strdup(path);
    prg_path[ptr-path]='\0';
    *prg = parse_prg(prg_path);
    free(prg_path);

    /* parse_prg could not find progam */
    if(*prg<0)
	return pt_invalid;

    /* Program only */
    if(*ptr=='\0')
	return pt_program;

    /* Extract track number, check it is present */
    sscanf(ptr,"/track-%d.wav", trk);
    if(*trk<0 || *trk>=24 || !fdmsfs.prgs[*prg].tracks[*trk].present)
	return pt_invalid;

    return pt_track;
}

static int fdmsfs_getattr(const char *path, struct stat *stbuf)
{
    enum fdmsfs_pathtype path_type;
    int prg, trk;

    memset(stbuf, 0, sizeof(struct stat));

    path_type = parse_path(path, &prg, &trk);

    switch(path_type) {
    case pt_root:
    case pt_program:
	stbuf->st_mode = S_IFDIR | 0555;
	stbuf->st_nlink = 2;
	break;

    case pt_track:
	/* Fill stat struct */
	stbuf->st_mode = S_IFREG | 0644;
	stbuf->st_nlink = 1;
	stbuf->st_size = sizeof(wavheader) +
	    fdmsfs.prgs[prg].tracks[trk].length*0x200LL *
	    fdmsfs.output_size/fdmsfs.sample_size;
	stbuf->st_blksize = 512;
	stbuf->st_blocks = stbuf->st_size / 512;
	break;

    default:
	return -ENOENT;
    }

    return 0;
}

static int fdmsfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi)
{
    (void) offset;
    (void) fi;
    enum fdmsfs_pathtype path_type;
    int prg,trk;

    path_type = parse_path(path, &prg, &trk);

    /* Program listing */
    switch(path_type) {
    case pt_root:
	filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);
	for(prg=0; prg<128; prg++) {
	    char filename[80];
	    if(!fdmsfs.prgs[prg].present)
		continue;
	    sprintf(filename, "p%.3d-%s", prg, fdmsfs.prgs[prg].name);
	    fdmsfs_clobber(filename);
	    filler(buf, filename, NULL, 0);
	    dprintf("Readdir: %s\n",filename);
	}
	break;

    case pt_program:
	filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);
	for(trk=0; trk<24; trk++) {
	    char filename[80];
	    sprintf(filename, "track-%d.wav",trk);
	    if(fdmsfs.prgs[prg].tracks[trk].present)
		filler(buf, filename, NULL, 0);
	}
	break;

    default:
	return -ENOENT;
    }

    return 0;
}

static int fdmsfs_open(const char *path, struct fuse_file_info *fi)
{
    struct fdmsfs_fh *fh;

    fprintf(stderr,"Open: %s\n",path);

    if((fi->flags & 0x3) != O_RDONLY)
	return -EACCES;

    fh = malloc(sizeof(struct fdmsfs_fh));

    if(NULL==fh)
	return -ENOMEM;

    memset(fh,0,sizeof(struct fdmsfs_fh));

    fh->pos = -1;
    fi->fh = (typeof(fi->fh)) fh;

    /* Parse path and store program/tack numbers */
    if(parse_path(path, &fh->prg, &fh->trk) != pt_track)
	return -ENOENT;

    init_segment_table(fh->prg, fh->trk);

    return 0;
}

static int fdmsfs_release(const char *path, struct fuse_file_info *fi)
{
    struct fdmsfs_fh *fh;

    fprintf(stderr,"Release: %s\n",path);

    fh = (struct fdmsfs_fh *) fi->fh;

    /* If a track, free segment table */
    if(parse_path(path, &fh->prg, &fh->trk) == pt_track)
	free_segment_table(fh->prg, fh->trk);

    free(fh);
    fi->fh = (typeof(fi->fh)) NULL;

    return 0;
}

static void set_wav_header(int prg, int trk) {
    ((unsigned int *)wavheader)[1] =
	fdmsfs.prgs[prg].tracks[trk].length*0x200LL*fdmsfs.output_size/fdmsfs.sample_size - 8;
    ((unsigned int *)wavheader)[6] = fdmsfs.sample_rate;
    ((unsigned int *)wavheader)[7] = fdmsfs.sample_rate*fdmsfs.output_size;
    ((unsigned int *)wavheader)[8] =
	fdmsfs.output_size | (fdmsfs.output_size<<19); //*8 then shift into second short
    ((unsigned int *)wavheader)[10] =
	fdmsfs.prgs[prg].tracks[trk].length*0x200LL*fdmsfs.output_size/fdmsfs.sample_size -
	sizeof(wavheader);
}

static int read_wav_header(char *output_buffer, size_t size, off_t offset) {
    /* Truncate to header length */
    if(offset>=sizeof(wavheader))
	return 0;
    if(offset+size>sizeof(wavheader))
	size = sizeof(wavheader) - offset;

    memcpy(output_buffer, wavheader + offset, size);

    return size;
}

static int read_samples(char *sample_buffer, size_t sample_size, off_t sample_offset, struct fdmsfs_fh *fh) {
    int s;

    /* Read samples */
    for(s=0;s<0x200;s++){
	size_t ssize;
	unsigned long s_offset;
	unsigned long s_length;

	s_offset = fdmsfs.prgs[fh->prg].tracks[fh->trk].segments[s].offset;
	s_length = fdmsfs.prgs[fh->prg].tracks[fh->trk].segments[s].length;

	/* Skip over segments until sample_offset falls within it */
	if(sample_offset >= s_length*0x200LL) {
	    sample_offset-= s_length*0x200LL;
	    continue;
	}

	/* Start with full read */
	ssize = sample_size;
	/* Trunc at end of segment */
	if(sample_offset+ssize > 0x200LL*s_length)
	    ssize = 0x200LL*s_length - sample_offset;

	if(s_offset) {
	    dprintf( "reading from segment @%lx\n", s_offset );
	    lseek(fdmsfs.fd, s_offset*0x200LL+sample_offset, SEEK_SET);

	    read(fdmsfs.fd, sample_buffer, ssize);
	} else {
	    memset(sample_buffer,0,ssize);
	}

	sample_offset = 0;
	sample_buffer += ssize;
	sample_size -= ssize;
    }

    return 0;
}

static inline void convert_s24le_s32le(char *out, char*in, int count) {
    do {
	out[4*count+0] = 0;
	out[4*count+1] = in[3*count+0];
	out[4*count+2] = in[3*count+1];
	out[4*count+3] = in[3*count+2];
    } while(count--);
}

static inline void convert_s16be_s16le(char *out, char*in, int count) {
    do {
	*( ((uint16_t *)out)+count ) = *( ((uint16_t *)in)+count );
	SWAP_16( ((uint16_t *)out)+count );
    } while(count--);
}

static int fdmsfs_read(const char *path, char *output_buffer, size_t size, off_t offset, struct fuse_file_info *fi)
{
    struct fdmsfs_fh *fh = (struct fdmsfs_fh *) fi->fh;
    void (*convert_sample) (char *, char *, int) = NULL;
    int s, sample_count;

    off_t sample_offset;
    size_t header_size, sample_size;
    off_t tmp_off;
    size_t tmp_size;

    int align_start, align_end;

    char *sample_buffer = NULL;


    /* Ensure wav header is correct.
     * Done here just in case a device ever supports multiple bit widths or
     * sample rates on different programs (or even tracks).
     */
    set_wav_header(fh->prg, fh->trk);

    /* Read wav header into output buffer */
    header_size = read_wav_header(output_buffer, size, offset);

    /* Subtract size of wavheader from offset, truncating to 0 */
    sample_offset = (offset<sizeof(wavheader)) ? 0 : ( offset - sizeof(wavheader) );
    /* Subtract read byte count from desired count */
    sample_size = size - header_size;

    /* Calculate sample offset and size.
     * The input offset and size may not be sample-aligned,
     * so we round the offset down to the nearest sample
     * and the size up to the nearest sample.
     * This lets us work with whole samples.
     * After conversion we can shift the output.
     *
     * Note that we need to read full samples because of sample
     * conversion.
     */

    tmp_off = sample_offset;
    sample_offset = ALIGN_DOWN(sample_offset, fdmsfs.output_size);
    align_start = tmp_off-sample_offset;
    sample_offset /= fdmsfs.output_size;
    sample_offset *= fdmsfs.sample_size;

    tmp_size = sample_size;
    sample_size = ALIGN_UP(sample_size, fdmsfs.output_size);
    align_end = sample_size-tmp_size;
    sample_size /= fdmsfs.output_size;
    sample_size *= fdmsfs.sample_size;

    /* Trying to read beyond end of track - Truncate */
    if(sample_offset>=fdmsfs.prgs[fh->prg].tracks[fh->trk].length*0x200LL)
	return 0;
    if(sample_offset+sample_size>fdmsfs.prgs[fh->prg].tracks[fh->trk].length*0x200LL)
	sample_size = fdmsfs.prgs[fh->prg].tracks[fh->trk].length*0x200LL - sample_offset;

    sample_buffer = malloc(sample_size);
    assert(sample_buffer!=NULL);

    if(read_samples(sample_buffer, sample_size, sample_offset, fh)) {
	free(sample_buffer);
	return -EIO;
    }

    /* Setup conversion function */
    switch(fdmsfs.sample_format) {
    case S16_BE:
	convert_sample = convert_s16be_s16le;
	break;
    case S24_LE:
	convert_sample = convert_s24le_s32le;
	break;
    default:
	assert(0);
    }

    /* Convert partial first sample */
    if(align_start) {
	char sample[4];

	convert_sample(sample, sample_buffer, 1);
	memcpy(output_buffer+header_size,
	       sample+align_start,
	       fdmsfs.sample_size-align_start);
    }

    s = align_start ? 1 : 0;
    sample_count =  sample_size/fdmsfs.sample_size - 
	(align_end ? 2 : 1);

    /* Convert full samples */
    convert_sample(output_buffer+header_size+s*fdmsfs.output_size,
		   sample_buffer+s*fdmsfs.sample_size, sample_count);

    if(align_end) {
	/* Convert partial last sample */
	char sample[4];
	s = sample_size/fdmsfs.sample_size;

	convert_sample(sample, sample_buffer+sample_size-fdmsfs.sample_size, 1);
	memcpy(output_buffer+header_size+s*fdmsfs.output_size,
	       sample+align_end,
	       fdmsfs.output_size-align_end);
    }

    free(sample_buffer);

    return size;
}

static struct fuse_operations fdmsfs_oper = {
    .getattr = fdmsfs_getattr,
    .readdir = fdmsfs_readdir,
    .open = fdmsfs_open,
    .read = fdmsfs_read,
    .release = fdmsfs_release,
};

static void usage()
{
    fprintf(stderr, "fdmsfs [--bits 24|16] [--rate r] <device|image> <mountpoint>\n");
}

off_t dev_size(int fd) {
    struct stat dev_stat;
    off_t size = 0;

    fstat(fd,&dev_stat);
    size = dev_stat.st_size;

#ifdef BLKGETSIZE
    /* If BLKGETSIZE (Linux) ioctl is available, use it */
    if(!size) {
	unsigned long numblocks = 0;
	if( -1 == ioctl(fd, BLKGETSIZE, &numblocks) ) {
	    perror("ioctl");
	    return 0;
	}
	size = numblocks * 512LL;
    }
#endif
#ifdef DKIOCGETBLOCKCOUNT
    /* If DKIOCGETBLOCKCOUNT (Mac OS X) ioctl is available, use it */
    if(!size) {
	unsigned long numblocks = 0;
	if( -1 == ioctl(fd, DKIOCGETBLOCKCOUNT, &numblocks) ) {
	    perror("ioctl");
	    return 0;
	}
	size = numblocks * 512LL;
    }
#endif

    /* Fall back on lseek-ing to end of file */
    if(!size) {
	if( -1 == (size = lseek(fd, 0, SEEK_END)) ) {
	    perror("lseek");
	    return 0;
	}
	if( -1 == lseek(fd, 0, SEEK_SET) ) {
	    perror("lseek");
	    return 0;
	}
    }

    return size;
}

int init(char *hdd)
{
    int i,j;

    fdmsfs.fd = open(hdd,O_RDONLY);
    fdmsfs.dev_size = dev_size(fdmsfs.fd);
    dprintf("Device size: %llu\n",fdmsfs.dev_size);

    /* Seek to directory */
    if(lseek(fdmsfs.fd, 0x201*0x200LL, SEEK_SET)<0){
	perror("Error seeking in file");
	exit(1);
    }

    /* Read directory */
    for(i=0; i<128; i++) {
	uint32_t offset;

	if(read(fdmsfs.fd,&offset,4)<0){
	    perror("Error reading file");
	    exit(1);
	}

	SWAP_32(&offset)

	    fdmsfs.prgs[i].offset=offset;
    }

    /* Read program headers */
    for(i=0; i<128; i++) {
	if(fdmsfs.prgs[i].offset==0)
	    continue;
	if(lseek(fdmsfs.fd, fdmsfs.prgs[i].offset*0x200LL, SEEK_SET)<0){
	    perror("Error seeking in file");
	    exit(1);

	}
	if(lseek(fdmsfs.fd, 72, SEEK_CUR)<0){
	    perror("Error seeking in file");
	    exit(1);
	}

	if(read(fdmsfs.fd,fdmsfs.prgs[i].name,16)<0){
	    perror("Error reading file");
	    exit(1);
	}

	j=16;
	while(fdmsfs.prgs[i].name[--j]==0x20) fdmsfs.prgs[i].name[j]='\0';

	/*
	 * (at least) D2424 can have null names.
	 * Need to figure out if offset==0 check is enough.
	 */
	//if(fdmsfs.prgs[i].name[0]=='\0')
	//    continue;

	fdmsfs.prg_count++;

	init_tracks(i);

	fdmsfs.prgs[i].present=0;
	for(j=0; j<24; j++)
	    if(fdmsfs.prgs[i].tracks[j].present)
		fdmsfs.prgs[i].present=1;

	if(verbose && fdmsfs.prgs[i].present)
	    fprintf(stderr,"Program: %s\n",fdmsfs.prgs[i].name);
    }

    return fdmsfs.fd;
}

void set_bits(char *arg, int usr)
{
    int bits = atoi(arg);
    switch(bits) {
    case 16:
	if(usr)
	    fprintf(stderr,"S16_BE forced\n");
	fdmsfs.sample_format = S16_BE;
	fdmsfs.sample_size = 2;
	fdmsfs.output_size = 2;
	break;
    case 24:
	if(usr)
	    fprintf(stderr,"S24_LE forced\n");
	fdmsfs.sample_format = S24_LE;
	fdmsfs.sample_size = 3;
	fdmsfs.output_size = 4;
	break;
    default:
	fprintf(stderr,"Invalid sample format: %s\n",arg);
	exit(1);
    }
}

void set_rate(char *arg)
{
    unsigned long rate = strtoul(arg,NULL,10);
    switch(rate) {
    case 44100:
    case 48000:
	break;
    default:
	fprintf(stderr,"WARNING: non-standard sample rate: %lu\n",rate);
    }
    fdmsfs.sample_rate = rate;
}

int main(int argc, char *argv[])
{
    char *hdd = NULL;
    char c;
    int nonopt_ind;
    int long_index;

    memset(&fdmsfs,0,sizeof(fdmsfs));

    /* Defaults */
    set_bits("16",0);
    set_rate("44100");

    /*Parse commandline*/
    opterr = 0;
    while((c = getopt_long(argc, argv, "hv", long_options, &long_index)) != -1) {
	switch (c) {
	case 'h':
	    usage();
	    return 0;
	case 'v':
	    verbose = 1;
	    break;
	case 0:
	    switch(lopt_flag) {
	    case LOPT_BITS:
		set_bits(optarg,1);
		break;
	    case LOPT_RATE:
		set_rate(optarg);
		break;
	    default:
		fprintf(stderr, "Unrecognized long option\n");
	    }
	    break;
	case '?':
	    fprintf(stderr, "Unknown option or option needs parameter: %s\n",
		    argv[optind-1]);
	    return 1;
	default:
	    abort();
	}
    }
    nonopt_ind=optind;

    if(nonopt_ind<argc) {
	hdd = strdup(argv[nonopt_ind]);
	nonopt_ind++;
    }

    if(hdd==NULL) {
	usage();
	return 0;
    }

    if(init(hdd) < 0) {
	fprintf(stderr,"Can't open fostex hdd device or image.\n");
	return 1;
    }

    fprintf(stderr,"Allocated %zu bytes\n",sizeof(fdmsfs));
    fprintf(stderr,"Found %d programs\n",fdmsfs.prg_count);
    fprintf(stderr,"Sample rate: %lu, Sample size (in): %d, Sample size (out): %d\n",
	    fdmsfs.sample_rate,fdmsfs.sample_size,fdmsfs.output_size);

    argv[--nonopt_ind] = argv[0];

    return fuse_main(argc-optind, argv+optind, &fdmsfs_oper);
}
